-- Emilio Francesquini <e.francesquini@ufabc.edu.br>
-- CC-BY-SA
-- 09/2019

{-# LANGUAGE BangPatterns, OverloadedStrings #-}
module Main where

import Control.Monad

import Data.List

import Data.Text (Text)
import qualified Data.Text as T

import Data.Text.Metrics

import Data.Vector (Vector)
import qualified Data.Vector as V

import Data.Set (Set)
import qualified Data.Set as S

import System.Directory (doesFileExist)

import System.IO

-- Cada palavra é representada como um vértice de um grafo
data Vertex = Vertex {
  idf             :: !Int, -- ID da palavra
  word            :: !Text,
  neighborIndices :: Set Int -- Cjto de IDs das palavras vizinhas
}

-- Grafo é representado como um vetor de vértices (para indexação
-- O(1))
newtype Graph = Graph (Vector Vertex)

-- Um caminho nada mais é do que uma lista de vertices
type Path = [Vertex]


-- Devolve o vértice de ID i
(!) :: Graph -> Int -> Vertex
(!) (Graph v) i = v V.! i

instance Eq Vertex where
  (Vertex ia wa _) == (Vertex ib wb _)
    | ia == -1 || ib == -1 = wa == wb
    | otherwise = ia == ib

instance Ord Vertex where
  compare a b = compare (word a) (word b)

neighbors :: Graph -> Vertex -> Set Vertex
neighbors g v = S.map (g !) (neighborIndices v)


loadPristineSetOfWords :: FilePath -> IO Graph
loadPristineSetOfWords fp = do
  g0 <- startUp fp
  return $ Graph (initGraph g0)
  where
    startUp fname = do
      fContents <- readFile fname -- uma palavra por linha
      return $ V.fromList $ zipWith l2t [0..] (sort $ lines fContents)
      where
        l2t i w = Vertex i (T.toUpper $ T.pack w) S.empty
    -- Como calcular a distância de levenshtein é caro, só calcula se
    -- tiver chance de ser 1
    conexo (Vertex _ p1 _) (Vertex _ p2 _) =
      abs (T.length p1 - T.length p2) < 2 && levenshtein p1 p2 == 1
    -- Dá para acelerar o processo se ordenarmos as palavras por
    -- tamanho e parar quando a diferença entre os tamanhos for > 1
    -- (em vez de ficar testando as diferenças de tamanho a cada
    -- comparação)
    initGraph g =
      V.map (\v@(Vertex i p _) -> Vertex i p (edges v)) g
      where
        edges v = S.fromList $ V.toList (V.map idf (V.filter (conexo v) g))

loadPreCalculatedGraph :: FilePath -> IO Graph
loadPreCalculatedGraph fname = do
  fContents <- readFile fname
  let fLines = lines fContents
      g =  Graph $ V.fromList $ zipWith (buildV g) [0..] (tail fLines)
  return g
  where
    buildV g i ln =
      Vertex i x (S.fromList (map (intBinSearch g) xs))
      where
        (x:xs) = map T.pack (words ln)


instance Show Vertex where
  show (Vertex i p v) =
       show i ++ ":\""
    ++ T.unpack p ++ "\" "
    ++ "(" ++ show (length v) ++ ")"

pathToText :: Path -> Text
pathToText [] = "Path not found"
pathToText path =
  T.intercalate " -> " mPathWs
  where
    mPathWs = map word path

vector :: Graph -> Vector Vertex
vector (Graph v) = v

size :: Graph -> Int
size (Graph v) = V.length v



intBinSearch :: Graph -> Text -> Int
intBinSearch g w =
  intBinSearch' 0 (size g)
  where
    intBinSearch' l r
      | r < l = -1
      | otherwise =
        case compare (g ! mid) v of
          GT -> intBinSearch' l (mid - 1)
          LT -> intBinSearch' (mid + 1) r
          EQ -> mid
        where
          v = Vertex (-1) w S.empty
          mid = (l + r) `div` 2

binSearch ::  Graph -> Text -> Maybe Vertex
binSearch g w =
  let res = intBinSearch g w in
    if res == -1
    then Nothing
    else Just $ g ! res

findWord :: Graph -> Text -> Maybe Vertex
findWord g w = binSearch g (T.toUpper w)

findPath :: Graph -> Vertex -> Vertex -> Maybe Path
findPath = bfs

-- Efetua uma busca em largura no grafo
bfs :: Graph -> Vertex -> Vertex -> Maybe Path
bfs g v0 v1 =
  map (g !) <$> mpath
  where
    s = idf v0
    d = idf v1
    mpath = bfs' [(s, [])] (S.singleton s)
    -- Fila (vertice, backtrack) -> Ver. Vistos
    bfs' :: [(Int, [Int])] -> Set Int -> Maybe [Int]
    bfs' [] _ = Nothing
    bfs' ((v, path):qs) !seen
      | d == v = Just $ reverse (d:path)
      | otherwise =
        bfs' nqueue nseen
        where
          fneigs = neighborIndices $ g ! v
          neigs = fneigs S.\\ seen
          nqueue = qs ++ zip (S.toList neigs) (repeat (v:path))
          nseen = S.union seen neigs

saveCache :: FilePath -> Graph -> IO ()
saveCache fp g =
  writeFile fp dump
  where
    vToStr v = unwords $ T.unpack (word v) : map (T.unpack . word) (S.toList $ neighbors g v)
    dump = unlines $ show (size g) : map vToStr (V.toList $ vector g)

loadFrom :: FilePath -> IO Graph
loadFrom fp = do
  putStrLn $ "Reading from: " <> fp
  isCached <- doesFileExist cachefp
  putStr $ "Cached file (" <> cachefp <> ") "
  if isCached
  then do
    putStrLn "found."
    loadPreCalculatedGraph cachefp
  else do
    putStrLn "not found."
    putStrLn "Pre-processing file. This will take a while..."
    g <- loadPristineSetOfWords fp
    putStrLn "Saving cache."
    saveCache cachefp g
    return g
  where
    cachefp = fp <> ".cache"

repl :: Graph -> IO ()
repl grafo = do
  origem <- prompt "Origem: "
  let mOrigem = findWord grafo $ T.pack origem
  destino <- prompt "Destino: "
  let mDestino =  findWord grafo $ T.pack destino
  putStrLn $ maybe
      "Caminho não encontrado"
      (T.unpack . pathToText)
      $ join $ findPath grafo <$> mOrigem <*> mDestino
  repl grafo

prompt :: String -> IO String
prompt what = do
    putStr what
    hFlush stdout
    getLine

main :: IO()
main = do
  fname <- prompt "Digite o nome do arquivo dicionario: "
  grafo <- loadFrom fname
  repl grafo

-- aura ausente
-- asa assalto
